#include <windows.h>
#include <stdio.h>



/*****************************************************************
 *                                                               *
 * The postnntp code, source and everything, is hereby placed in *
 * the Public Domain. That means you may do with it whatever you *
 * wish, including any alterations, selling, or leasing it. The  *
 * only action you may not take is claim copyright on property   *
 * placed in the Public Domain.                                  *
 *                                                               *
 *                             - Felix Kasza (felixk@dms.shd.de) *
 *                                                               *
 *****************************************************************/



// postnntp.cpp: reads a file as a stream of RFC-1036 messages.
// The articles are posted to an NNTP server specified on the
// command line. No validation is performed on articles. An
// article's end is reached at the end-of-file, or when a line
// with only a dot "." on it is read, whichever comes first. To
// include a line beginning with a "." in the article text,
// double the "."



// constants
#define	MAXL	2048 // twice the NNTP legal limit



// these are options and the like
int opt_debugnntp = 0; // debug output if non-0
const char *opt_hostname = "localhost"; // this machine
const char *opt_proto = "nntp"; // service to use
int opt_reader = 0; // do "mode reader" if non-0
int opt_authn = 0; // do authentication if non-0
const char *opt_user = ""; // username for auth'n
const char *opt_pass = ""; // password for auth'n
int opt_single = 0; // single-article mode


// global vars? Who, _me_?
SOCKET sock; // the socket over which we talk



// function prototypes
int main( int argc, char *argv[] );
void wserr( int rc, const char * const funcname );
int r( char *p ); // read a line from the other guy
void w( const char *const p, int ignoredbg = 0 ); // write a line to the other guy
void postfile( FILE *fp ); // posts the contents of a file
int getl( FILE *fp, char *p ); // read a line from a file



// getl() reads a line and returns -1 if end of file,
// 0 otherwise
int getl( FILE *fp, char *p )
{
	int c, cnt = 0;

	while ( ( c = getc( fp ) ) != EOF && c != '\n' && c != '\r' )
	{
		if ( c == '\r' )
			continue; // gobble up CRs
		*( p ++ ) = (char) c;
		cnt ++;
		// check line length
		if ( cnt >= MAXL - 2 ) // reserve 1 for '\0', one for doubled dot
			break;
	}
	*p = '\0';

	return c == EOF? EOF: 0;
}



// postfile() breaks a file into articles and posts them
void postfile( FILE *fp )
{
	// states for the state machine:
	// 0 -- before article, try to get a line
	// 2 -- in article, read till "." or EOF comes along
	// 3 -- in article, but read till EOF, doubling dots
	// 8 -- at end of article, terminate it
	// 9 -- not really a state, EOF was hit
	// transitions:
	// 0 -> 9 if EOF
	// 0 -> 2 if message line && !opt_single
	// 0 -> 3 if message line && opt_single
	// 2 -> 2 if !EOF && !dot
	// 2 -> 8 if dot || EOF
	// 3 -> 3 if !EOF
	// 3 -> 8 if EOF
	// 8 -> 0 if !EOF
	// 8 -> 9 if EOF

	int rc, eof, state = 0; // POST not yet sent
	char l[MAXL], l2[MAXL], *p;

	while ( 1 )
	{
		eof = getl( fp, l );
		if ( eof && *l != '\0' ) // did we get a line?
			eof = 0; // clear, it will get set next time around
reevaluate:
		switch ( state )
		{
		case 0: // not in an article
			if ( eof )
			{
				state = 9;
				goto reevaluate;
			}
			for ( p = l; *p > '\0' && *p <= ' '; p ++ )
				; // skip whitespace
			if ( *p == '\0' ) // nothing left?
				break; // go get next line
			w( "post" );
			rc = r( l2 ); // read response
			if ( rc < 300 || rc > 399 )
			{
				fprintf( stderr, "POST command failed: '%s'\n", l2 );
				return;
			}
			w( l, 1 ); // send current line
			// we apparently got a line, next state is 1 or 2
			state = opt_single? 3: 2;
			break;
		case 2: // in article, wait for eof or dot
			if ( eof || ( *l == '.' && l[1] == '\0' ) )
			{ // above we made sure that if ( end ) then current line is empty or dot
				state = 8;
				goto reevaluate; // post this article
			}
			// normal article line
			w( l, 1 );
			// remain in this state
			state = 2;
			break; // go through the loop again
		case 3: // in article, wait for eof
			if ( eof )
			{ // above we made sure that if ( end ) then current line is empty or dot
				state = 8;
				goto reevaluate; // post this article
			}
			// check for leading dot
			if ( l[0] == '.' )
			{
				memmove( l + 1, l, strlen( l ) + 1 );
				*l = '.';
			}
			// normal article line
			w( l, 1 );
			// remain in this state
			state = 3;
			break; // go through the loop again
		case 8: // at end of article, finish it
			w( ".", 1 ); // terminate article
			rc = r( l2 );
			if ( rc < 200 || rc > 299 ) // error?
				fprintf( stderr, "POST failed, continuing: '%s'\n", l2 );
			state = eof? 9: 0;
			break;
		case 9: // at eof, quit
			return;
		}
	}
}



// rdc() returns the next character from the socket.
char rdc()
{
	static char buf[MAXL];
	static int nbuf = 0; // chars in buffer
	static char *bufp; // pointer to next byte to return

	while ( nbuf == 0 ) // must fill buffer?
	{
		bufp = buf;
		nbuf = recv( sock, buf, sizeof buf, 0 );
		if ( nbuf == SOCKET_ERROR )
			wserr( nbuf, "recv" );
	}

	nbuf --;
	return *( bufp ++ );
}



// r() reads a line from the socket and returns the value of the
// integer at its start
int r( char *p )
{
	char c;
	char *orig_p = p;

	// gobble up \r, \n
	while ( ( c = rdc() ) == '\r' || c == '\n' )
		;
	// the first byte
	*( p ++ ) = c;
	while ( ( c = rdc() ) != '\r' && c != '\n' )
		*( p ++ ) = c;

	*p = '\0'; // terminate received string

	if ( opt_debugnntp )
	{
		fputs( " -> ", stderr );
		fputs( orig_p, stderr );
		fputs( "\n", stderr );
	}

	return atoi( orig_p );
}



// w() writes a string to the other guy and terminates it with
// carriage-return./linefeed. ignoredbg, if true, suppresses debug output for article lines.
void w( const char *const p, int ignoredbg /* = 0 */ )
{
	static char crlf[] = "\r\n";

	if ( strlen( p ) != (size_t) send( sock, p, strlen( p ), 0 ) )
		wserr( 3, "send" );
	if ( 2 != send( sock, crlf, 2, 0 ) )
		wserr( 4, "send" );

	if ( opt_debugnntp )
	{
		if ( !ignoredbg )
		{
			fputs( "<-  ", stderr );
			fputs( p, stderr );
			fputs( "\n", stderr );
		}
		else
		{
			if ( *p == '.' && p[1] == '\0' ) // article end?
				fputs( "<-  [article]\n", stderr );
		}
	}
}



// wserr() displays winsock errors and aborts. No grace there.
void wserr( int rc, const char * const funcname )
{
	if ( rc == 0 )
		return;

	fprintf( stderr, "\nWinsock error %d [%d] returned by %s().\n"
		"Sorry, no bonus!\n", rc, WSAGetLastError(), funcname );
	WSACleanup();
	exit( rc );
}



// initialize() handles banners, "mode reader", and authentication
void initialize()
{
	char line[MAXL];
	int rc;

	// read the banner line
	rc = r( line );
	if ( rc != 200 )
	{
		fprintf( stderr, "Bad banner: '%s'\n", line );
		exit( 2 );
	}

	if ( opt_reader )
	{
		// send mode reader
		w( "mode reader" );

		// read banner
		rc = r( line );
		if ( rc != 200 )
		{
			fprintf( stderr, "Bad banner: '%s'\n", line );
			exit( 2 );
		}
	}

	if ( opt_authn )
	{
		char cmd[MAXL];

		// send authinfo user
		sprintf( cmd, "authinfo user %s", opt_user );
		w( cmd );
		// wait for 381
		rc = r( line );
		if ( rc != 381 )
		{
			fprintf( stderr, "Authentication (user) failed: '%s'\n", line );
			exit( 2 );
		}

		// send authinfo pass
		sprintf( cmd, "authinfo pass %s", opt_pass );
		w( cmd );
		// wait for 2xx
		rc = r( line );
		if ( rc < 200 || rc > 299 )
		{
			fprintf( stderr, "Authentication (password) failed: '%s'\n", line );
			exit( 2 );
		}
	}
}



int main( int argc, char *argv[] )
{
	int rc, i, usage = 0, port;
	char l[MAXL];
	unsigned long naddr;
	FILE *fp;
	WSADATA wsadata;
	PHOSTENT phe;
	PSERVENT pse;
	SOCKADDR_IN addr;

	// parse switches
	if ( argc == 1 )
	{
show_usage:
		fputs( "postnntp [options] [file ...]\n\n", stderr );
		fputs( "Options are valid for the entire command. Supported options:\n", stderr );
		fputs( "-d: output all sends and receives on the socket\n", stderr );
		fputs( "-s <server>: the NNTP server's address, either a DNS name or\n", stderr );
		fputs( "    an IP address in dotted-decimal notation. Default: 127.0.0.1\n", stderr );
		fputs( "-p <port>: The port to talk to on the server, as a decimal\n", stderr );
		fputs( "    number or a service as per getservbyname(). Default: 119\n", stderr );
		fputs( "-u <username> <password>: Requires two args, a username and a\n", stderr );
		fputs( "    password, and causes authentication commands to be sent. For\n", stderr );
		fputs( "    now, only clear-text authentication is supported.\n", stderr );
		fputs( "-r: causes a \"mode reader\" command to be sent as the very first\n", stderr );
		fputs( "    thing after connecting.\n", stderr );
		fputs( "-z: single-article mode. Only EOF marks the end of the article.\n", stderr );
		fputs( "    Dots at line beginnings will automatically be doubled.\n", stderr );
		fputs( "file: input files, no wildcards, as many as you like. \"-\" causes\n", stderr );
		fputs( "    postnntp to read from stdin.\n\n", stderr );
		fputs( "postnntp reads a file as a stream of RFC-1036 messages.\n", stderr );
		fputs( "The articles are posted to an NNTP server specified on the\n", stderr );
		fputs( "command line. No validation is performed on articles. An\n", stderr );
		fputs( "article's end is reached at the end-of-file, or when a line\n", stderr );
		fputs( "with only a dot \".\" on it is read, whichever comes first. To\n", stderr );
		fputs( "include a line beginning with a \".\" in the article text,\n", stderr );
		fputs( "double the \".\"\n", stderr );
		return 1;
	}

	for ( i = 1; i < argc; i ++ )
	{
		if ( *argv[i] != '-' || argv[i][1] == '\0' )
			continue;
		switch ( argv[i][1] )
		{
		case '?':
			goto show_usage;
		case 'd': // debug
			opt_debugnntp ++;
			break;
		case 'r': // debug
			opt_reader ++;
			break;
		case 'z': // debug
			opt_single ++;
			break;
		case 's': // server
			if ( i >= argc - 1 )
			{
				fputs( "-s requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_hostname = argv[++ i];
			break;
		case 'p': // port
			if ( i >= argc - 1 )
			{
				fputs( "-p requires an argument.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_proto = argv[++ i];
			break;
		case 'u': // username, password
			if ( i >= argc - 2 )
			{
				fputs( "-u requires two arguments.\n\n", stderr );
				usage = 1;
				break;
			}
			argv[i] = NULL;
			opt_user = argv[++ i];
			argv[i] = NULL;
			opt_pass = argv[++ i];
			opt_authn = 1;
			break;
		default:
			fprintf( stderr, "Unknown option: '%s'\n", argv[i] );
			usage = 1;
			break;
		}
		argv[i] = NULL;
	}
	if ( usage )
		goto show_usage;

	rc = WSAStartup( 2, &wsadata );
	wserr( rc, "WSAStartup" );

	sock = socket( AF_INET, SOCK_STREAM, 0 );
	if ( sock == INVALID_SOCKET )
		wserr( 999, "socket" );

	addr.sin_family = AF_INET;
	// try numeric IP address first (inet_addr)
	naddr = inet_addr( opt_hostname );
	if ( naddr != INADDR_NONE )
	{
		addr.sin_addr.s_addr = naddr;
	}
	else
	{
		phe = gethostbyname( opt_hostname );
		if ( phe == NULL )
			wserr( 1, "gethostbyname" );
		addr.sin_addr.s_addr = *( (unsigned long *) (phe->h_addr) );
		memcpy( (char *) &addr.sin_addr, phe->h_addr, phe->h_length );
	}

	// try numeric protocol first
	port = atoi( opt_proto );
	if ( port > 0 && port < 23768 )
		addr.sin_port = htons( (short) port );
	else
	{
		pse = getservbyname( opt_proto, "tcp" );
		if ( pse == NULL )
			wserr( 1, "getservbyname" );
		addr.sin_port = pse->s_port;
	}

	rc = connect( sock, (SOCKADDR *) &addr, sizeof addr );
	wserr( rc, "connect" );

	// initialize, authenticate, and so on
	initialize();

	// process input
	for ( i = 1; i < argc; i ++ )
	{
		if ( argv[i] != NULL )
		{
			if ( *argv[i] == '-' && argv[i][1] == '\0' ) // stdin?
				postfile( stdin );
			else
			{
				fp = fopen( argv[i], "rt" );
				if ( fp == NULL )
					fprintf( stderr, "cannot open file, skipping: '%s'\n", argv[i] );
				else
				{
					postfile( fp );
					fclose( fp );
				}
			}
		}
	}

	// close connection with QUIT
	w( "quit" );
	rc = r( l );
	if ( rc < 200 || rc > 299 )
		fprintf( stderr, "Bad response to QUIT: '%s'\n", l );

	// close socket
	rc = closesocket( sock );
	wserr( rc, "closesocket" );

	rc = WSACleanup();
	wserr( rc, "WSACleanup" );

	return 0;
}
